@@ -6,17 +6,19 @@ import sys
 import time
 import weakref
 
-from greenhouse import compat
+from greenhouse import compat, poller
 
 
 __all__ = ["pause", "pause_until", "pause_for", "schedule", "schedule_at",
         "schedule_in", "schedule_recurring", "schedule_exception",
         "schedule_exception_at", "schedule_exception_in", "end",
-        "global_exception_handler", "local_exception_handler",
-        "handle_exception", "greenlet", "global_trace_hook",
-        "local_incoming_trace_hook", "local_outgoing_trace_hook"]
+        "global_exception_handler", "remove_global_exception_handler",
+        "local_exception_handler", "remove_local_exception_handler",
+        "handle_exception", "greenlet", "global_hook", "remove_global_hook",
+        "local_incoming_hook", "remove_local_incoming_hook",
+        "local_outgoing_hook", "remove_local_outgoing_hook",
+        "set_ignore_interrupts", "reset_poller"]
 
-POLL_TIMEOUT = 1.0
 BTREE_ORDER = 64
 
 
@@ -36,19 +38,19 @@ state.to_run = collections.deque()
 
 # exceptions queued up for scheduled coros
 state.to_raise = weakref.WeakKeyDictionary()
-state.raise_in_main = None
 
 # exception handlers, global and local
 state.global_exception_handlers = []
 state.local_exception_handlers = weakref.WeakKeyDictionary()
 
 # trace hook callbacks
-state.global_trace_hooks = []
-state.local_to_trace_hooks = weakref.WeakKeyDictionary()
-state.local_from_trace_hooks = weakref.WeakKeyDictionary()
+state.global_hooks = []
+state.local_to_hooks = weakref.WeakKeyDictionary()
+state.local_from_hooks = weakref.WeakKeyDictionary()
 
 # tracks interrupts
 state.interrupted = False
+state.ignore_interrupts = False
 
 
 class TimeoutManager(object):
@@ -125,20 +127,27 @@ except ImportError:
 
 
 def _hit_poller(timeout):
-    until = time.time() + timeout
     events = []
     while 1:
         try:
             events = state.poller.poll(timeout)
             break
+        except KeyboardInterrupt, exc:
+            # on Ctrl-C, wake up the main without killing the mainloop
+            state.to_raise[compat.main_greenlet] = exc
+            state.to_run.append(compat.main_greenlet)
+            return
         except EnvironmentError, exc:
             if exc.args[0] != errno.EINTR:
                 raise
+
             # interrupted by a signal
-            # wake them all up and let them know there was an interruption
-            state.interrupted = True
-            events = [(fd, state.poller.ERRMASK)
-                    for fd in state.poller._registry.iterkeys()]
+            if state.ignore_interrupts:
+                events = []
+            else:
+                state.interrupted = True
+                events = [(fd, state.poller.ERRMASK)
+                        for fd in state.poller._registry.iterkeys()]
             break
 
     for fd, eventmap in events:
@@ -166,8 +175,13 @@ def _hit_poller(timeout):
             for writable in writables:
                 writable()
 
-    _check_events()
-    _check_paused()
+    state.to_run.extend(state.awoken_from_events)
+    state.awoken_from_events.clear()
+
+    state.timed_paused.check()
+
+    state.to_run.extend(state.paused)
+    state.paused = []
 
 
 class _WeakMethodRef(object):
@@ -205,16 +219,6 @@ def _register_fd(fd, readable, writable):
     state.descriptormap[fd].append((readable, writable))
 
 
-def _check_events():
-    state.to_run.extend(state.awoken_from_events)
-    state.awoken_from_events.clear()
-
-def _check_paused():
-    state.timed_paused.check()
-
-    state.to_run.extend(state.paused)
-    state.paused = []
-
 def greenlet(func, args=(), kwargs=None):
     """create a new greenlet from a function and arguments
 
@@ -445,9 +449,7 @@ def schedule_exception(exception, target):
     :param target: the greenlet that should receive the exception
     :type target: greenlet
     """
-    if target is compat.main_greenlet:
-        state.raise_in_main = exception
-    elif not isinstance(target, compat.greenlet):
+    if not isinstance(target, compat.greenlet):
         raise TypeError("can only schedule exceptions for greenlets")
     if target.dead:
         raise ValueError("can't send exceptions to a dead greenlet")
@@ -464,8 +466,6 @@ def schedule_exception_at(unixtime, exception, target):
     :param target: the greenlet that should receive the exception
     :type target: greenlet
     """
-    if target is compat.main_greenlet:
-        state.raise_in_main = exception
     if not isinstance(target, compat.greenlet):
         raise TypeError("can only schedule exceptions for greenlets")
     if target.dead:
@@ -491,8 +491,6 @@ def end(target):
     :param target: the greenlet to end
     :type target: greenlet
     """
-    if target is compat.main_greenlet:
-        state.raise_in_main = compat.GreenletExit()
     if not isinstance(target, compat.greenlet):
         raise TypeError("argument must be a greenlet")
     if not target.dead:
@@ -530,29 +528,24 @@ def mainloop():
                 until = state.timed_paused.first()[0] + 0.001
                 _hit_poller(until - time.time())
             else:
-                _hit_poller(POLL_TIMEOUT)
+                _hit_poller(None)
 
         prev = target
         target = state.to_run.popleft()
 
         # global trace hooks
-        if state.global_trace_hooks:
-            _run_global_trace_hooks(prev, target)
+        if state.global_hooks:
+            _run_global_hooks(prev, target)
 
         # local trace incoming hooks
-        if target in state.local_to_trace_hooks:
-            _run_local_trace_hooks(
-                    target, state.local_to_trace_hooks[target], True)
+        if target in state.local_to_hooks:
+            _run_local_hooks(
+                    target, state.local_to_hooks[target], True)
 
         try:
             # pick up any exception we are supposed to throw in
             if target in state.to_raise:
                 target.throw(state.to_raise.pop(target))
-            elif (target is compat.main_greenlet
-                    and state.raise_in_main is not None):
-                exc = state.raise_in_main
-                state.raise_in_main = None
-                target.throw(exc)
             else:
                 target.switch()
         except Exception:
@@ -564,13 +557,13 @@ def mainloop():
             del klass, exc, tb
 
         # local trace outgoing hooks
-        if target in state.local_from_trace_hooks:
-            _run_local_trace_hooks(
-                    target, state.local_from_trace_hooks[target], False)
+        if target in state.local_from_hooks:
+            _run_local_hooks(
+                    target, state.local_from_hooks[target], False)
 
 state.mainloop = mainloop
 
-def _run_local_trace_hooks(target, hooks, incoming):
+def _run_local_hooks(target, hooks, incoming):
     replacement_hooks = []
     direction = 1 if incoming else 2
     for weak in hooks:
@@ -587,9 +580,9 @@ def _run_local_trace_hooks(target, hooks, incoming):
 
     hooks[:] = replacement_hooks
 
-def _run_global_trace_hooks(coming_from, going_to):
+def _run_global_hooks(coming_from, going_to):
     replacement_hooks = []
-    for weak in state.global_trace_hooks:
+    for weak in state.global_hooks:
         func = weak()
         if func is None:
             continue
@@ -601,7 +594,7 @@ def _run_global_trace_hooks(coming_from, going_to):
 
         replacement_hooks.append(weak)
 
-    state.global_trace_hooks[:] = replacement_hooks
+    state.global_hooks[:] = replacement_hooks
 
 
 def handle_exception(klass, exc, tb, coro=None):
@@ -620,6 +613,10 @@ def handle_exception(klass, exc, tb, coro=None):
         behave as though the exception occurred in this coroutine (defaults to
         the current coroutine)
     :type coro: greenlet
+
+    exception handlers run would be all those added with
+    :func:`global_exception_handler`, and any added for the relevant coroutine
+    with :func:`local_exception_handler`.
     """
     if coro is None:
         coro = compat.getcurrent()
@@ -659,9 +656,11 @@ def global_exception_handler(handler):
     """add a callback for when an exception goes uncaught in any greenlet
 
     :param handler:
-        the callback function. must be a function taking 3 arguments: ``klass``
-        the exception class, ``exc`` the exception instance, and ``tb`` the
-        traceback object
+        the callback function. must be a function taking 3 arguments:
+
+        - ``klass`` the exception class
+        - ``exc`` the exception instance
+        - ``tb`` the traceback object
     :type handler: function
 
     Note also that the callback is only held by a weakref, so if all other refs
@@ -674,12 +673,32 @@ def global_exception_handler(handler):
 
     return handler
 
+def remove_global_exception_handler(handler):
+    """remove a callback from the list of global exception handlers
+
+    :param handler:
+        the callback, previously added via :func:`global_exception_handler`,
+        to remove
+    :type handler: function
+
+    :returns: bool, whether the handler was found (and therefore removed)
+    """
+    for i, cb in enumerate(state.global_exception_handlers):
+        cb = cb()
+        if cb is not None and cb is handler:
+            state.global_exception_handlers.pop(i)
+            return True
+    return False
+
 def local_exception_handler(handler=None, coro=None):
     """add a callback for when an exception occurs in a particular greenlet
 
     :param handler:
-        the callback function, must be a function taking 3 arguments: the
-        exception class, the exception instance, and the traceback object
+        the callback function, must be a function taking 3 arguments:
+
+        - ``klass`` the exception class
+        - ``exc`` the exception instance
+        - ``tb`` the traceback object
     :type handler: function
     :param coro:
         the coroutine for which to apply the exception handler (defaults to the
@@ -700,37 +719,84 @@ def local_exception_handler(handler=None, coro=None):
 
     return handler
 
-def global_trace_hook(handler):
+def remove_local_exception_handler(handler, coro=None):
+    """remove a callback from the list of exception handlers for a coroutine
+
+    :param handler: the callback to remove
+    :type handler: function
+    :param coro: the coroutine for which to remove the local handler
+    :type coro: greenlet
+
+    :returns: bool, whether the handler was found (and therefore removed)
+    """
+    if coro is None:
+        coro = compat.getcurrent()
+
+    for i, cb in enumerate(state.local_exception_handlers.get(coro, [])):
+        cb = cb()
+        if cb is not None and cb is handler:
+            state.local_exception_handlers[coro].pop(i)
+            return True
+    return False
+
+
+def global_hook(handler):
     """add a callback to run in every switch between coroutines
 
     :param handler:
-        the callback function, must be a function taking 2 arguments: the
-        greenlet being switched from, and the greenlet being switched to. be
-        aware that only a weak reference to this function will be held.
+        the callback function, must be a function taking 2 arguments:
+
+        - the greenlet being switched from
+        - the greenlet being switched to
+
+        be aware that only a weak reference to this function will be held.
     :type handler: function
     """
     if not hasattr(handler, "__call__"):
         raise TypeError("trace hooks must be callable")
 
-    state.global_trace_hooks.append(weakref.ref(handler))
+    state.global_hooks.append(weakref.ref(handler))
 
     return handler
 
-def local_incoming_trace_hook(handler=None, coro=None):
+def remove_global_hook(handler):
+    """remove a callback from the list of global hooks
+
+    :param handler:
+        the callback function, previously added with global_hook, to remove
+        from the list of global hooks
+    :type handler: function
+
+    :returns: bool, whether the handler was removed from the global hooks
+    """
+    for i, cb in enumerate(state.global_hooks):
+        cb = cb()
+        if cb is not None and cb is handler:
+            state.global_hooks.pop(i)
+            return True
+    return False
+
+
+def local_incoming_hook(handler=None, coro=None):
     """add a callback to run every time a greenlet is about to be switched to
 
     :param handler:
-        the callback function, must be a function taking 2 arguments: an
-        integer indicating whether it is being called as an incoming (1) hook
-        or as an outgoing (2) hook. in this case it will always be 1. be aware
-        that only a weak reference to this function will be held.
+        the callback function, must be a function taking 2 arguments:
+
+        - an integer indicating whether it is being called as an incoming (1)
+          hook or an outgoing (2) hook (in this case it will always receive 1).
+        - the coroutine being switched into (in this case it will always be the
+          same as the one indicated by the ``coro`` argument to
+          ``local_incoming_hook``.
+
+        Be aware that only a weak reference to this function will be held.
     :type handler: function
     :param coro:
         the coroutine for which to apply the trace hook (defaults to current)
     :type coro: greenlet
     """
     if handler is None:
-        return lambda h: local_incoming_trace_hook(h, coro)
+        return lambda h: local_incoming_hook(h, coro)
 
     if not hasattr(handler, "__call__"):
         raise TypeError("trace hooks must be callable")
@@ -738,26 +804,50 @@ def local_incoming_trace_hook(handler=None, coro=None):
     if coro is None:
         coro = compat.getcurrent()
 
-    state.local_to_trace_hooks.setdefault(coro, []).append(
+    state.local_to_hooks.setdefault(coro, []).append(
             weakref.ref(handler))
 
     return handler
 
-def local_outgoing_trace_hook(handler=None, coro=None):
+def remove_local_incoming_hook(handler, coro=None):
+    """remove a callback from the incoming hooks for a particular coro
+
+    :param handler: the callback previously added via local_incoming_hook
+    :type handler: function
+    :param coro: the coroutine for which the hook should be removed
+    :type coro: greenlet
+
+    :returns: bool, whether the handler was found and removed
+    """
+    if coro is None:
+        coro = compat.getcurrent()
+
+    for i, cb in enumerate(state.local_to_hooks.get(coro, [])):
+        cb = cb()
+        if cb is not None and cb is handler:
+            state.local_to_hooks[coro].pop(i)
+            return True
+    return False
+
+def local_outgoing_hook(handler=None, coro=None):
     """add a callback to run every time a greenlet is switched away from
 
     :param handler:
-        the callback function, must be a function taking 2 arguments: an
-        integer indicating whether it is being called as an incoming (1) hook
-        or as an outgoing (2) hook. in this case it will always be 2. be aware
-        that only a weak reference to this function will be held.
+        the callback function, must be a function taking 2 arguments:
+
+        - an integer indicating whether it is being called as an incoming (1)
+          hook or as an outgoing (2) hook (in this case it will always be 2).
+        - the coroutine being switched from (in this case it is the one
+          indicated by the ``coro`` argument to ``local_outgoing_hook``.
+
+        Be aware that only a weak reference to this function will be held.
     :type handler: function
     :param coro:
         the coroutine for which to apply the trace hook (defaults to current)
     :type coro: greenlet
     """
     if handler is None:
-        return lambda h: local_outgoing_trace_hook(h, coro)
+        return lambda h: local_outgoing_hook(h, coro)
 
     if not hasattr(handler, "__call__"):
         raise TypeError("trace hooks must be callable")
@@ -765,7 +855,48 @@ def local_outgoing_trace_hook(handler=None, coro=None):
     if coro is None:
         coro = compat.getcurrent()
 
-    state.local_from_trace_hooks.setdefault(coro, []).append(
+    state.local_from_hooks.setdefault(coro, []).append(
             weakref.ref(handler))
 
     return handler
+
+def remove_local_outgoing_hook(handler, coro=None):
+    """remove a callback from the outgoing hooks for a particular coro
+
+    :param handler: the callback previously added via local_outgoing_hook
+    :type handler: function
+    :param coro: the coroutine for which the hook should be removed
+    :type coro: greenlet
+
+    :returns: bool, whether the handler was found and removed
+    """
+    if coro is None:
+        coro = compat.getcurrent()
+
+    for i, cb in enumerate(state.local_from_hooks.get(coro, [])):
+        cb = cb()
+        if cb is not None and cb is handler:
+            state.local_from_hooks[coro].pop(i)
+            return True
+    return False
+
+def set_ignore_interrupts(flag=True):
+    """turn off EINTR-raising from emulated syscalls on interruption by signals
+
+    due to the nature of greenhouse's system call emulation,
+    ``signal.siginterrupt`` can't be made to work with it. specifically,
+    greenhouse can't differentiate between different signals. so this function
+    toggles whether to restart for *ALL* or *NO* signals.
+
+    :param flag:
+        whether to turn EINTR exceptions off (``True``) or on (``False``)
+    :type flag: bool
+    """
+    state.ignore_interrupts = bool(flag)
+
+def reset_poller(poll=None):
+    """replace the scheduler's poller, throwing away any pre-existing state
+
+    this is only really a good idea in the new child process after a fork(2).
+    """
+    state.poller = poll or poller.best()
